/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

/* psf analysis and extraction */

#define _GNU_SOURCE
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <math.h>
#include <ctype.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "gcx.h"
#include "catalogs.h"
#include "gui.h"
#include "interface.h"
#include "params.h"
#include "sourcesdraw.h"
#include "obsdata.h"
#include "recipy.h"
#include "symbols.h"
#include "wcs.h"
#include "multiband.h"
#include "filegui.h"
#include "plots.h"
#include "psf.h"
#include "dsimplex.h"

#define GROWTH_POINTS 128
#define growth_radius(s) (1.0 + 0.5 * (s))

int growth_curve(struct ccd_frame *fr, double x, double y, double grc[], int n)
{
	struct ap_params apdef;
	struct star s;
	int i, ret;
	double mflux = 1;

	ap_params_from_par(&apdef);

	memset(&s, 0, sizeof(struct star));
	s.x = x;
	s.y = y;
	s.xerr = BIG_ERR;
	s.yerr = BIG_ERR;
	if (P_INT(AP_AUTO_CENTER)) {
		center_star(fr, &s, P_DBL(AP_MAX_CENTER_ERR));
	}
	apdef.center = P_INT(AP_AUTO_CENTER);
	ret = aperture_photometry(fr, &s, &apdef, NULL);
	if (ret)
		return ret;
	apdef.center = 0;

	for (i = 0; i < n; i++) {
		apdef.r1 = growth_radius(i);
		ret = aperture_photometry(fr, &s, &apdef, NULL);
		if (ret)
			return ret;
		grc[i] = s.aph.tflux - s.aph.star_all * s.aph.sky;
		d4_printf("r:%.1f tflux:%.1f star_all:%.3f\n", 
			  apdef.r1, s.aph.tflux, s.aph.star_all);
		if (apdef.r1 <= P_DBL(AP_R1))
			mflux = grc[i];
	}
	for (i = 0; i < n; i++) {
		grc[i] = grc[i] / mflux;
	}	
	return 0;
}

/* fill up the rp_point table with radius/intensity pairs. return the number of points filled */
int radial_profile(struct ccd_frame *fr, double x, double y, double r, 
		   struct rp_point rpp[], int n, double *peak, double *flux, 
		   double *sky, double *err)
{
	struct ap_params apdef;
	struct star s;
	int ret, ns = 0;
	float *d, *de;
	int ri, i;
	int ix, iy, xs, ys, xe, ye;
	double max = 1.0;

	ap_params_from_par(&apdef);

	memset(&s, 0, sizeof(struct star));
	s.x = x;
	s.y = y;
	s.xerr = BIG_ERR;
	s.yerr = BIG_ERR;
	if (P_INT(AP_AUTO_CENTER)) {
		center_star(fr, &s, P_DBL(AP_MAX_CENTER_ERR));
		apdef.center = P_INT(AP_AUTO_CENTER);
	}
	ret = aperture_photometry(fr, &s, &apdef, NULL);
	if (ret)
		return ret;

	ri = 2 * ceil(r);
	xs = floor(x - r);
	ys = floor(y - r);
	xe = xs + ri;
	ye = ys + ri;
	d = fr->dat;
	d += fr->w * ys + xs;
	de = fr->dat;
	de += fr->w * fr->h - 1;
	g_return_val_if_fail(ri < fr->w, 0);
	for (iy = 0; iy < ri; iy++) {
		for (ix = 0; ix < ri; ix++) {
			double dist;
			dist = sqrt(sqr(ix + xs - x) + sqr(iy + ys - y));
			if (ns < n && dist <= r) {
//				d4_printf("dist %8.1f v:%.1f\n", dist, *d);
				rpp[ns].r = dist;
				rpp[ns].v = *d;
				if (dist < r/2 && max < *d)
					max = *d;
				ns++;
			}
			d++;
			if (d >= de)
				break;
		}
		d += fr->w - ri;
		if (d >= de)
			break;
	}
	for (i = 0; i < ns; i++) {
		if ((max - s.aph.sky) / max > 0.001)
			rpp[i].v = (rpp[i].v - s.aph.sky) / (max - s.aph.sky);
		else 
			rpp[i].v = (rpp[i].v) / (max);
	}
	if (peak)
		*peak = max;
	if (sky)
		*sky = s.aph.sky;
	if (flux)
		*flux = s.aph.tflux - s.aph.star_all * s.aph.sky;
	if (err)
		*err = sqrt(sqr(s.aph.flux_err) + sqr(s.aph.star_all * s.aph.sky_err));
	if (ns < n) {
		rpp[ns].v = -1;
		rpp[ns].r = -1;
		return ns;
	} else {
		rpp[ns].v = -1;
		rpp[ns].r = -1;
		return ns - 1;
	}
}


int do_plot_profile(struct ccd_frame *fr, GSList *selection)
{
	GSList *sl, *vsl;
	struct gui_star *gs;
	int n, i, j;
	FILE *dfp = NULL;
	GSList *gr = NULL;
	double *grc = NULL;
	struct rp_point *rpp = NULL;
	int nrp;
	GSList *vs = NULL;
	char name[256];
	char preamble[16384];
	int np=0, nn = 0, pop;
	double peak, sky, flux;

	g_return_val_if_fail(fr != NULL, -1);

	n =  2 * GROWTH_POINTS * P_DBL(AP_R1) / growth_radius(GROWTH_POINTS) + 0.5;
	if (n > GROWTH_POINTS)
		n = GROWTH_POINTS;

	nrp = sqr(4 * ceil(P_DBL(AP_R1)));
	
	np += snprintf(preamble+np, 16384-np, "set key below\n");
	np += snprintf(preamble+np, 16384-np, "set y2tics autofreq\n");
	np += snprintf(preamble+np, 16384-np, "set y2label 'encircled flux'\n");
	np += snprintf(preamble+np, 16384-np, "set ylabel 'radial profile'\n");
	np += snprintf(preamble+np, 16384-np, "set grid xtics\n");
	np += snprintf(preamble+np, 16384-np, "plot [0:%.1f] 0 notitle lt 0, 1 axes x1y2 notitle lt 0, ",
		       2 * ceil(P_DBL(AP_R1)) );
	i = 0;
	
	for (sl = selection; sl != NULL; sl = sl->next) {
		double A, s;
		gs = GUI_STAR(sl->data);
		if (grc == NULL) {
			grc = malloc(GROWTH_POINTS * sizeof(double));
		}
		if (rpp == NULL) {
			rpp = malloc(nrp * sizeof(struct rp_point));
		}
		if (growth_curve(fr, gs->x, gs->y, grc, n))
			continue;
		if (radial_profile(fr, gs->x, gs->y, 2 * P_DBL(AP_R1), rpp, nrp,
				   &peak, &flux, &sky, NULL) <= 0)
			continue;
		A = P_DBL(AP_R1) / FWHMSIG / 2; s = 1.0;
		fit_1d_profile(rpp, &A, NULL, &s, NULL);
		if (gs->s)
			nn = snprintf(name, 255, "%s ", 
				      CAT_STAR(gs->s)->name);
		else
			nn = snprintf(name, 255, "x%.0fy%.0f ", gs->x, gs->y);
		snprintf(name+nn, 255-nn, "(FWHM:%.1f peak:%.0f sky:%.1f)",
			 FWHMSIG * s, peak, sky);
		for (j=0; rpp[j].r >= 0; j++) {
			rpp[j].v /= A;
		}
		
		gr = g_slist_append(gr, grc);
		vs = g_slist_append(vs, rpp);
		grc = NULL;
		rpp = NULL;
		if (i > 0)
			np += snprintf(preamble+np, 16384-np,", ");
		np += snprintf(preamble+np, 16384-np,"'-' title '%s' lt %d, "
			"exp(-0.5*x*x/%.2f) notitle lt %d, "
			"'-' axes x1y2 notitle with lines lt %d",
			name, i+1, s*s, i+1, i+1);
		i++;
	}
	if (g_slist_length(gr) <= 0) {
		if (grc)
			free(grc);
		if (rpp)
			free(rpp);
		err_printf("Bad star (too close to edge?)\n");
		return -1;
	}

	pop = open_plot(&dfp, NULL);
	if (pop >= 0) {
		fprintf(dfp, "%s\n", preamble);

		for (sl = gr, vsl = vs; sl != NULL && vsl != NULL; 
		     sl = sl->next, vsl = vsl->next) {
			struct rp_point *rp;
			rp = vsl->data;
			for(i = 0; rp[i].r >= 0; i++)
				fprintf(dfp, "%.2f %.3f\n", rp[i].r, rp[i].v);
			fprintf(dfp, "e\n");
			for(i=0; i<n; i++)
				fprintf(dfp, "%.2f %.3f\n", growth_radius(i), 
					((double *)sl->data)[i]);
			fprintf(dfp, "e\n");
		}
		close_plot(dfp, pop);
	}

	if (grc)
		free(grc);
	if (rpp)
		free(rpp);
	for (sl = gr; sl != NULL; sl = sl->next) {
		free(sl->data);
	}	
	for (sl = vs; sl != NULL; sl = sl->next) {
		free(sl->data);
	}	
	g_slist_free(gr);
	g_slist_free(vs);
	return 0;
}

/* minimisation error function (gaussian model, looking for peak and sigma).
   p[1] is the peak, p[2] is sigma, data is a radial profile. The error is 
   only calculated up to three sigmas. the profiles end with an entry with 
   negative radius */
double gauss_1d_2_err_funk(double p[], void *rppi)
{
	double A = p[1];
	double s = p[2];
	int i;
	double err = 0;
	struct rp_point *rpp = rppi;

	if (s < 0.01)
		return 1000;
	
	i = 0;
	while (rpp[i].r >= 0) {
			err += sqr(A * exp(-0.5 * sqr((rpp[i].r) / s)) - rpp[i].v);
		i++;
	}
//	d4_printf("A:%.3f s:%.3f err:%.3f\n", A, s, err);
	return err;
}

/* fit an unidimensional gaussian or moffat profile. if B is NULL, 
   the background is assumed 0 (not fitted). if b is NULL, a gaussian 
   is fitted, else a moffat function. */
int fit_1d_profile(struct rp_point *rpp, double *A, double *B, double *s, double *b)
{
	double **p;		/* np+1 rows, np columns (+1 due to 1-based) */
	double *y;		/* np columns ( " ) */
	int iter=0;
	int i;
	int ret;
	int np;

	np = 2;

	g_return_val_if_fail(A != NULL, -1);
	g_return_val_if_fail(s != NULL, -1);

	p = (double **) malloc((np + 2) * sizeof(double *));
	y = (double *) malloc((np + 2) * sizeof(double));

	for (i = 1; i <= np + 1; i++) {
		p[i] = (double *) malloc((np + 1) * sizeof(double));
	}

	p[1][1] = *A;
	p[1][2] = *s;

	p[2][1] = *A * 1.1;
	p[2][2] = *s;

	p[3][1] = *A;
	p[3][2] = *s * 1.1;

	for (i = 1; i <= np+1; i++) {
		y[i] = gauss_1d_2_err_funk(p[i], rpp);
	}

	ret = amoeba(p, y, np, 0.01, gauss_1d_2_err_funk, &iter, rpp);

	d3_printf("it:%d evals:%d s:%.3f A:%.3f\n", iter, ret, p[1][2], p[1][1]);

	*A = p[1][1];
	*s = p[1][2];

	for (i = 1; i < np + 2; i++)
		free((void *) p[i]);
	free((void *) p);
	free((void *) y);

	return ret;
}

void print_star_measures(gpointer window, GSList *found) 
{
	struct gui_star *gs;
	struct rp_point *rpp = NULL;
	int nrp;
	char name[256];
	double peak, sky, flux, err;
	struct image_channel *i_ch;
	double A, s;

	g_return_if_fail(window != NULL);

	i_ch = gtk_object_get_data(GTK_OBJECT(window), "i_channel");
	if (i_ch == NULL || i_ch->fr == NULL) {
		err_printf_sb2(window, "No image\n");
		error_beep();
		return;
	}

	if (found == NULL)
		return;
	gs = GUI_STAR(found->data);

	nrp = sqr(4 * ceil(P_DBL(AP_R1)));
	rpp = malloc(nrp * sizeof(struct rp_point));
	
	if (radial_profile(i_ch->fr, gs->x, gs->y, 2 * P_DBL(AP_R1), rpp, nrp,
			   &peak, &flux, &sky, &err) <= 0) {
		err_printf_sb2(window, "Bad star (too close to edge?)\n");
		error_beep();
		return;
	}
	A = P_DBL(AP_R1) / FWHMSIG / 2; s = 1.0;
	fit_1d_profile(rpp, &A, NULL, &s, NULL);
	if (gs->s)
		snprintf(name, 255, "%s %.1f,%.1f", 
			      CAT_STAR(gs->s)->name, gs->x, gs->y);
	else
		snprintf(name, 255, "%.1f,%.1f", gs->x, gs->y);
	info_printf_sb2(window, "%s  FWHM:%.1f peak:%.0f sky:%.1f flux:%.0f mag:%.03f/%.3f",
			name, FWHMSIG * s, peak, sky, flux, flux_to_absmag(flux), 
			fabs(flux_to_absmag(fabs(flux) + fabs(err)) - flux_to_absmag(fabs(flux))));
	free(rpp);

}
